/*
This Routine calculates one Runga-Kutta step over an interval from x to x+h

-> Inputs: <-

double y[]     Array of values for ODE variables specified in recursive relationships 
              at begining of interval
double dxdy[]  Array of derivitive values for ODE variables specified in recursive relationships 
              at begining of interval
int n         Number of recursive relationships, limit of indecies of y[],dxdy[],yout[],yerr

double x       Start of interval

double h       Duration of interval

void (*derivs)(double, double[], double[], double[])  function pointer to recursive relationship calculator                

<- Outputs: ->

double yout[]  Array of values for ODE variables specified in recursive relationships 
              at end of interval                                
double yerr[]  Array of estimates of error for recurive relationships


|  Notes: |

This routine assumes the values for the ODE variables and their derivatives are supplied for
the start of the interval in y and dxdy respectivly

The Runga-Kutta Method is based on Euler approximated integration. It is based on the principle
that an ODE (ordinary differential Equation) can be reduced to a set of first order differential 
equations. This method is used to solve intitial value problems for the ODE, where the only
constraint conditions are at the begining of an interval. This is not the most computationally
efficient method, predictor corrector methods may yeild better results, however RK is easy to 
implement and is virtually always succesful in finding a solution.

Solution of an ODE begins by decomposing it into a group of first order DEs
Following this, the set of FODEs (first order DE) must be translated into a recursive 
algebraic relationships representing the RHS of the FODEs, and relating them to the integral of 
the LHS of other FODEs in the set.

Following this, this RK method can be used in conjunction with a function to evaluate the RHS of the
algebraic equations for the set of FODEs to solve the ODE.

example:   Pendulum Equation

(1) d2o(t)/dt2 + (g/L)*sin(o(t)) = 0  o(0) = 0 do(0)/dt=0 

where o(t) is the pendulum angle at t, d2o(t)/dt2 is it's second derivitive 
and do(t)/dt is it's first derivative.

let y1=o

let y2=dy1/dt=do/dt

substituting into (1)

(2) dy2/dt = -(g/L)*sin(y1) 
(3) dy1/dt = y2

thus void (*derivs)(double x, double y[], double dydx[]) must calculate 

dydx[0]=y[1]
                                                     
and

dydx[1]=-(g/L)*sin(y[0]), where y[0] = y1, y[1] = y2, dydx[0] = dy1/dt, dydx[1] = dy2/dt

This RK method will then estimate the integral of these recursive relationships over the interval
while estimating the trucation error over the interval  
 


This procedure is known as the Embedded Runga-Kutta Cash-Karp method. This is similar to a
fouth order RK method, but computes a fifth order RK and uses the extra terms to estimate error.

The charicteristic equations of the method are below

k1 = h*f(xn,yn)
k2 = h*f(xn+a2*h,yn+b21k1)
...
k6 = h*f(xn+a6*h,yn+b61*k1+...+b65*k5)

below y(n+1) denotes the value of y at n+1 subscript

y(n+1) = yn + c1*k1 + c2*k2 + c3*k3 + c4*k4 + c5*k5 + c6*k6

the embedded 4th order formula is

y(n+1)* = yn + dc1*k1 + dc2*k2 + dc3*k3 + dc4*k4 + dc5*k5 + dc6*k6

the error is estimated as y(n+1)-y(n+1)* = sum 1 to 6 of [(cn-dcn)kn]
 

Cash-Karp Parameters for Embedded Runga-Kutta Method
____________________________________________________________________________________________________
I  i  I  a    I                       bij                                  I  c       I   dc        I
I_____I_______I____________________________________________________________I__________I_____________I
I     I       I   j=1      I   j=2    I    j=3     I  j=4         I j=5    I          I             I
I_____I_______I____________I__________I____________I______________I________I__________I_____________I
I  1  I       I            I          I            I              I        I 37/378   I  2825/27648 I
I_____I_______I____________I__________I____________I______________I________I__________I_____________I
I  2  I  1/5  I  1/5       I          I            I              I        I 0        I  0          I
I_____I_______I____________I__________I____________I______________I________I__________I_____________I
I  3  I  3/10 I  3/40      I 9/40     I            I              I        I 250/621  I  18575/48384I
I_____I_______I____________I__________I____________I______________I________I__________I_____________I
I  4  I  3/5  I  3/10      I -9/10    I  6/5       I              I        I 125/594  I  13525/55296I
I_____I_______I____________I__________I____________I______________I________I__________I_____________I
I  5  I  1    I  -11/54    I 5/2      I  -70/27    I 35/27        I        I 0        I  277/14336  I
I_____I_______I____________I__________I____________I______________I________I__________I_____________I
I  6  I  7/8  I  1631/55296I 175/512  I  575/13824 I 44275/110592 I253/4096I 512/1771 I  1/4        I
I_____I_______I____________I__________I____________I______________I________I__________I_____________I

*/


#include <math.h>
#include "ODESolver.h"
#include <stdio.h>
#include <stdlib.h>

void rkck(double y[], double dydx[], int n, double x, double h, double yout[], double yerr[], void (*derivs)(double, double[], double[], double[]), double coupled_inputs[])
{
    int i;
    
    
    // Cash-Karp Parameters for Embedded Runga-Kutta Method
    static double a2=0.2, a3=0.3, a4=0.6, a5=1.0, a6=0.875, 
        b21=0.2, 
        b31=3.0/40.0, b32=9.0/40.0, 
        b41=0.3, b42=-0.9, b43=1.2, 
        b51=-11.0/54.0, b52=2.5, b53=-70.0/27.0, b54=35.0/27.0,
        b61=1631.0/55296.0, b62=175.0/512.0, b63=575.0/13824.0, b64=44275.0/110592.0, b65=253.0/4096.0, 
        c1=37.0/378.0, c3=250.0/621.0, c4=125.0/594.0, c6=512.0/1771.0;
        

    double dc1=c1-2825.0/27648.0, dc3=c3-18575.0/48384.0,
        dc4=c4-13525.0/55296.0, dc5=-277.00/14336.0, dc6=c6-0.25;

    double *ak2, *ak3, *ak4, *ak5, *ak6, *ytemp;

    //initialize pointers with memory pointers
    ak2=(double *)calloc(n,sizeof(double));
    ak3=(double *)calloc(n,sizeof(double));
    ak4=(double *)calloc(n,sizeof(double));
    ak5=(double *)calloc(n,sizeof(double));
    ak6=(double *)calloc(n,sizeof(double));
    ytemp=(double *)calloc(n,sizeof(double));

    //first RKCK step

    for(i=0;i<n;i++)
        ytemp[i]=y[i]+b21*h*dydx[i];
    

    //second RKCK step
    
    (*derivs)(x+a2*h,ytemp,ak2,coupled_inputs);

    for(i=0;i<n;i++)
        ytemp[i]=y[i]+h*(b31*dydx[i]+b32*ak2[i]);
    

    //third RKCK step

    (*derivs)(x+a3*h,ytemp,ak3,coupled_inputs);
    
    for(i=0;i<n;i++)
        ytemp[i]=y[i]+h*(b41*dydx[i]+b42*ak2[i]+b43*ak3[i]);
    
    //fourth RKCK step
    
    (*derivs)(x+a4*h,ytemp,ak4,coupled_inputs);

    for(i=0;i<n;i++)
        ytemp[i]=y[i]+h*(b51*dydx[i]+b52*ak2[i]+b53*ak3[i]+b54*ak4[i]);
    

    //fifth RKCK step

    (*derivs)(x+a5*h,ytemp,ak5,coupled_inputs);

    for(i=0;i<n;i++)
        ytemp[i]=y[i]+h*(b61*dydx[i]+b62*ak2[i]+b63*ak3[i]+b64*ak4[i]+b65*ak5[i]);
    

    //sixth RKCK step

    (*derivs)(x+a6*h,ytemp,ak6,coupled_inputs);

    for(i=0;i<n;i++)  //accumulate increments with proper weights
        yout[i]=y[i]+h*(c1*dydx[i]+c3*ak3[i]+c4*ak4[i]+c6*ak6[i]);

    for(i=0;i<n;i++)  //estimate error as difference between fourth and fifth order method
        yerr[i]=h*(dc1*dydx[i]+dc3*ak3[i]+dc4*ak4[i]+dc5*ak5[i]+dc6*ak6[i]);

    free(ytemp);
    free(ak6);
    free(ak5);
    free(ak4);
    free(ak3);
    free(ak2);
}


status rkstep(double y[], double dydx[], int n, double *x, double htry, double eps, double yscal[], double *hdid, 
            double *hnext, void (*derivs)(double, double[], double[], double[]), double coupled_inputs[])
{
    int i;
    double errmax, h, htemp, xnew, *yerr, *ytemp, errtemp, hhtemp;
    
    yerr=(double *)calloc(n,sizeof(double));
    ytemp=(double *)calloc(n,sizeof(double));

    h=htry;

    for(;;) 
    {
        rkck(y,dydx,n,*x,h,ytemp,yerr,derivs,coupled_inputs); //take step
        
        errmax=0.0;
        for(i=0;i<n;i++){
            if((errtemp=fabs(yerr[i]/yscal[i])) > errmax)
                errmax=errtemp; //find maximum error during step
        }
        
        #ifdef PrintConsole
        printf("errmax: %f\n",errmax);
        #endif
        
        errmax /=eps;                                   //scale relative to required tolerance 
        
        if(errmax<=1.0) break;                          //step success, exit loop!
        
        htemp=SAFTY*h*pow(errmax,PSHRNK);
            //error too large, reduce step size, never below 10%
        
        if(htemp>=(hhtemp = 0.1*h)) //setup h for next iteration
            h=htemp;
        else
            h=hhtemp;


        xnew=(*x)+h; 
        if(xnew==(*x)) 
        {
            fprintf(stderr,"underflow in rkstep\n"); //underflow in h
            return ERROR;
        }
    }
    
    if(errmax>ERRCON) 
        *hnext=SAFTY*h*pow(errmax,PGROW);
    else 
        *hnext=5.0*h;                                  // no more than a five fold increase in step
    
    *x+=(*hdid=h);  //increment x to next step point
    
    for(i=0;i<n;i++)  //copy ytemp to output
        y[i] = ytemp[i];
    
    free(ytemp);
    free(yerr);
    return OK;
}

status ODESolve(double ystart[], int nvar, double x1, double x2, double epsilon, double h1, double hmin,
         int *nok, int *nbad, void (*derivs)(double, double[], double[], double[]), double *xp, double **yp, 
         double dxsave, int *kount, int kmax, double coupled_inputs[])
{
    int nstp, i;
    double xsave, x, hnext, hdid, h;
    double *yscal, *y, *dydx;

    yscal=(double *)calloc(nvar,sizeof(double));
    y=(double *)calloc(nvar,sizeof(double));
    dydx=(double *)calloc(nvar,sizeof(double));

    x=x1;
    h=h1;
    *nok = (*nbad) = (*kount) = 0;

    for(i=0;i<nvar;i++) //copy initial values 
        y[i]=ystart[i];

    if(kmax) 
        xsave=x-dxsave*2.0; 

    for(nstp=1;nstp<=MAXSTP;nstp++) // iterate unless max step size 
    {
        #ifdef PrintConsole
        printf("%d\n",nstp);
        #endif

        (*derivs)(x,y,dydx,coupled_inputs);   //find initial derivitive for rkstep 
        
        for(i=0;i<nvar;i++)   // find scaler for error on interval
            yscal[i]=fabs(y[i])+fabs(dydx[i]*h)+TINY;
        
        #ifdef RecordResults
        if(kmax > 0 && (*kount) < kmax-1 && (x-xsave) >= dxsave) //store intermediate results
        {
            xp[++(*kount)]=x;
            for(i=0;i<nvar;i++)
                yp[i][(*kount)]=y[i];
            xsave=x;
        }
        #endif

        if((x+h-x2) > 0.0) // prevent overrunning x2
            h=x2-x;
        
        if(rkstep(y,dydx,nvar,&x,h,epsilon,yscal,&hdid,&hnext,derivs,coupled_inputs)!=OK) //take error controlled step
        {
            return ERROR;
        }


        #ifdef PrintConsole
            
            printf("%f, ",x);
            for(i=0;i<nvar;i++)
            {
                printf("%f, ",y[i]);
            }
            printf("\nhdid: %f hnext: %f x-x2: %f",hdid,hnext,x-x2);
            printf("\n");
        #endif


        if(hdid == h) // increment step counters 
            ++(*nok);
        else
            ++(*nbad);

        if((x-x2)>=0.0) //done interval from x1 to x2
        {
            for(i=0;i<nvar;i++) //copy final results for interval to ystart
                ystart[i]=y[i];
            
            #ifdef RecordResults
            if(kmax)             //copy last intermediate result
            {
                xp[++(*kount)]=x;
                for(i=0;i<nvar;i++)
                    yp[i][(*kount)]=y[i];

            }
            #endif

            free(dydx);
            free(y);
            free(yscal);
            return OK;
        }
        
        if (hnext<=hmin) 
        {
            printf("Step size below hmin in ODESolve\n");
            fprintf(stderr,"Step size below hmin in ODESolve\n");
            return ERROR;
        }
        
        h=hnext; // proceed to next step

    }
    
    printf("Too many steps in ODESolve\n");
    fprintf(stderr,"Too many steps in ODESolve\n");
    
    return ERROR;
}  
